﻿using System;
using System.Collections.Generic;
using System.Windows.Media.Media3D;
using ProjectorKit;
using ProjectorKit.Elements;
using ProjectorKit.Events;
using ProjectorKit.Filter;

namespace CalgaryMap
{
    public class Program
    {
        private const int MAX_HIGH_RES_DISTANCE = 1000;
        // Number of tiles
        private const int ROW_TILE_COUNT = 8;
        private const int COL_TILE_COUNT = 8;
        // Resulting map is 3010x3010mm
        private const int TILE_SIZE_X = 430;
        private const int TILE_SIZE_Y = 430;

        private const int ROWS_OUTSIDE_DISPLAY = 3;
        private const int COLS_OUTSIDE_DISPLAY = 3;

        private IEnvironment env = new XmlEnvironment();
        private List<string> highResTileNames = new List<string>();

        static void Main(string[] args)
        {
            new Program();
        }

        public Program()
        {
            AbstractDisplay smartboard = (ActiveDisplay)env.World.Get("Smartboard");

            // Add the tiled map to the world
            for (int row = 0; row < ROW_TILE_COUNT; row++)
            {
                for (int col = 0; col < COL_TILE_COUNT; col++)
                {
                    int xOff = (row - ROW_TILE_COUNT / 2) * TILE_SIZE_X;
                    int yOff = -(col - COL_TILE_COUNT / 2) * TILE_SIZE_Y;
                    Point3D pos = smartboard.Position + new Vector3D(xOff, yOff, 0);
                    ImageElement low = new ImageElement(pos, TILE_SIZE_X, TILE_SIZE_Y, @"map\low\tile_" + row + "_" + col + ".png");
                    env.World.Add(low);

                    if (row >= ROWS_OUTSIDE_DISPLAY && row <= ROW_TILE_COUNT - ROWS_OUTSIDE_DISPLAY && 
                        col >= COLS_OUTSIDE_DISPLAY && col <= COL_TILE_COUNT - COLS_OUTSIDE_DISPLAY)
                    {
                        string name = "tile_" + row + "_" + col;
                        ImageElement high = new ImageElement(pos + new Vector3D(0, 0, 1), TILE_SIZE_X, TILE_SIZE_Y, @"map\high\tile_" + row + "_" + col + ".png");
                        highResTileNames.Add(name);
                        env.World.Add(name, high);

                        // Hide high-res always on smartboard
                        smartboard.HideElement(name);
                    }
                }
            }

            // Create overlappings
            AbstractDisplay[] projectors =  {
                (AbstractDisplay)env.World.Get("Dell"),
                (AbstractDisplay)env.World.Get("Projector2")
            };
            foreach (AbstractDisplay projector in projectors)
            {
                projector.Filter = new LinearGridFilter(0.0005);

                OverlappingDisplaysEvent or = new OverlappingDisplaysEvent(smartboard, projector);
                or.OverlappingStarted += OverlappingCallback;
                or.OverlappingChanged += OverlappingCallback;
                or.OverlappingEnded += OverlappingEnd;
            }
        }

        private void OverlappingCallback(object sender, OverlappingDisplayEventArgs e)
        {
            if (e.DisplayArgs1.NumberOfOverlappingPixels >= e.DisplayArgs2.NumberOfOverlappingPixels)
            {
                // Smartboard has more pixels, only show smartboard
                e.Display1.RemoveBlackout(env.World.GetNameOf(e.Display2));
                e.Display2.AddBlackout(env.World.GetNameOf(e.Display1));

                // No High Resolution
                foreach (string name in highResTileNames)
                    e.Display2.HideElement(name);
            }
            else
            {
                // Projector has more pixels, clip smartboard
                // Calculate the opacity for a fade-in effect
                double opacity = 1;
                double difference = Math.Min(e.DisplayArgs1.NumberOfOverlappingPixels, e.DisplayArgs2.NumberOfOverlappingPixels) /
                    Math.Max(e.DisplayArgs1.NumberOfOverlappingPixels, e.DisplayArgs2.NumberOfOverlappingPixels);
                if (difference > 0.8)
                    opacity = (1 - difference) * 5;

                e.Display2.RemoveBlackout(env.World.GetNameOf(e.Display1));
                e.Display1.AddBlackout(env.World.GetNameOf(e.Display2), opacity);

                // High Resolution
                foreach (string name in highResTileNames)
                    e.Display2.ShowElement(name);
            }
        }

        private void OverlappingEnd(object sender, OverlappingDisplayEventArgs e)
        {
            // Remove clipping in both viewports
            e.Display1.RemoveBlackout(env.World.GetNameOf(e.Display2));
            e.Display2.RemoveBlackout(env.World.GetNameOf(e.Display1));

            // No High Resolution
            foreach (string name in highResTileNames)
                e.Display2.HideElement(name);
        }
    }
}
