﻿using System.Collections.Generic;
using System.Linq;
using System.Windows.Media.Media3D;
using ProjectorKit;
using ProjectorKit.Elements;
using ProjectorKit.World;

namespace Picture_Viewer_with_External_Buttons
{   /// <summary>
    /// Shows pictures on the smartboard in fullscreen. 
    /// Next, previous and delete buttons are off-screen around the display. 
    /// They can be revealed with the projector and clicked with a Wii Remote.
    /// </summary>
    class Program
    {
        private readonly int CONTROL_SIZE = 200;
        private IEnvironment env;
        private List<ImageElement> images = new List<ImageElement>();
        private int index = 0;
        public int Index
        {
            get { return index; }
            set
            {
                // Set value and bring image to foreground (focus)
                index = value;
                if (images.Count > 0)
                    ServerWorld.Instance.Focus(images.ElementAt(index));
            }
        }

        static void Main(string[] args)
        {
            new Program();
        }

        public Program()
        {
            // Create a new environment
            env = new XmlEnvironment();

            // Add images in fullscreen on the Smartboard
            ActiveDisplay sb = (ActiveDisplay)env.World.Get("Smartboard");
            for (int i = 1; i <= 4; i++)
            {
                ImageElement ie = new ImageElement(sb.Position, sb.Width, sb.Height, @"images\" + i + @".jpeg");
                images.Add(ie);
                env.World.Add(ie);
            }

            // Add controls to the borders of the smartboard
            Point3D bottom = new Point3D(sb.Position.X, sb.Position.Y - sb.Height / 2 - CONTROL_SIZE / 2, sb.Position.Z);
            Point3D left = new Point3D(sb.Position.X - sb.Width / 2 - CONTROL_SIZE / 2, sb.Position.Y, sb.Position.Z);
            Point3D right = new Point3D(sb.Position.X + sb.Width / 2 + CONTROL_SIZE / 2, sb.Position.Y, sb.Position.Z);
            env.World.Add(new ButtonElement(bottom, sb.Width, CONTROL_SIZE, @"images\trash.jpeg", Delete));
            env.World.Add(new ButtonElement(left, CONTROL_SIZE, sb.Height, @"images\previous.png", Previous));
            env.World.Add(new ButtonElement(right, CONTROL_SIZE, sb.Height, @"images\next.png", Next));

            // Use WiiRemote as controller, use A for clicking
            WiiRemoteElement wiiremote = new WiiRemoteElement();
            new ClickHandler(wiiremote.ButtonA);
            env.Tracker.BindToTrackedObject(wiiremote, "MarkerA");
            env.CursorDB.Register(wiiremote);

            // Focus image at the first index
            Index = 0;
        }

        /// <summary>
        /// Show previous image.
        /// </summary>
        /// <param name="element">Not needed</param>
        private void Previous(ButtonElement element)
        {
            Index = (index + (images.Count - 1)) % images.Count;
        }

        /// <summary>
        /// Show next image.
        /// </summary>
        /// <param name="element">Not needed</param>
        private void Next(ButtonElement element)
        {
            Index = (index + 1) % images.Count;
        }

        /// <summary>
        /// Delete current image.
        /// </summary>
        /// <param name="element">Not needed</param>
        private void Delete(ButtonElement element)
        {
            env.World.Remove(images.ElementAt(index));
            images.Remove(images.ElementAt(index));
            Previous(element);
        }
    }
}
