﻿using System;
using System.Collections.Generic;
using System.Windows.Media.Media3D;
using System.Xml;
using ProjectorKit.Elements;
using ProjectorKit.Tracking;
using ProjectorKit.Utilities;

namespace ProjectorKit.Configuration
{
    public class XmlConfiguration : IConfiguration
    {
        private Dictionary<string, AbstractElement> namedElements = new Dictionary<string, AbstractElement>();
        private List<AbstractElement> unnamedElements = new List<AbstractElement>();
        private List<TrackerBinding> trackerBindings = new List<TrackerBinding>();
        #region Properties
        public IDictionary<string, AbstractElement> NamedElements
        {
            get { return namedElements; }
        }

        public IEnumerable<AbstractElement> UnnamedElements
        {
            get { return unnamedElements; }
        }

        public IEnumerable<TrackerBinding> TrackerBindings
        {
            get { return trackerBindings; }
        }
        #endregion

        public XmlConfiguration(string file)
        {
            XmlDocument xmlDoc = new XmlDocument();
            try
            {
                xmlDoc.Load(file);
            }
            catch (Exception)
            {
                throw new Exception("The configuration file is not valid XML.");
            }

            LoadAllElements(xmlDoc);
            LoadAllTrackerBindings(xmlDoc);
        }

        private void LoadAllTrackerBindings(XmlDocument xmlDoc)
        {
            XmlNodeList bindings = xmlDoc.GetElementsByTagName("tracker_binding");
            foreach (XmlNode binding in bindings)
            {
                string marker = binding.Attributes["marker"].Value;
                string elementName = binding.Attributes["element_name"].Value;
                if (namedElements.ContainsKey(elementName))
                {
                    TrackerBinding tb;
                    if (binding["offset"] != null)
                    {
                        Vector3D offset = LoadVector(binding["offset"]);
                        tb = new TrackerBinding(marker, namedElements[elementName], offset);
                    }
                    else
                        tb = new TrackerBinding(marker, namedElements[elementName]);
                    trackerBindings.Add(tb);
                }
                else
                    throw new Exception("Configuration Error: Element with name '" + elementName + "' is not defined!");
            }
        }

        private IEnumerable<AbstractElement> LoadAllElements(XmlDocument xmlDoc)
        {
            List<AbstractElement> elements = new List<AbstractElement>();

            XmlNodeList walls = xmlDoc.GetElementsByTagName("wall");
            foreach (XmlNode node in walls)
                AddElement(node, LoadWall(node));

            XmlNodeList projectors = xmlDoc.GetElementsByTagName("projector");
            foreach (XmlNode node in projectors)
                AddElement(node, LoadProjector(node));

            XmlNodeList displays = xmlDoc.GetElementsByTagName("display");
            foreach (XmlNode node in displays)
                AddElement(node, LoadDisplay(node));

            return elements;
        }

        private void AddElement(XmlNode node, AbstractElement element)
        {
            if (node.Attributes["name"] != null)
            {
                string name = node.Attributes["name"].Value;
                if (namedElements.ContainsKey(name))
                    throw new Exception("Configuration Error: Multiple definitions of '" + name + "'.");
                else
                    namedElements[name] = element;
            }
            else
                unnamedElements.Add(element);
        }

        #region Load Elements
        private Projector LoadProjector(XmlNode node)
        {
            Projector projector;
            if (node.Attributes["center_y"] != null)
                projector = new Projector(
                    int.Parse(node.Attributes["resolution_x"].Value),
                    int.Parse(node.Attributes["resolution_y"].Value),
                    Angle.FromDegrees(int.Parse(node.Attributes["horizontal_angle"].Value)),
                    int.Parse(node.Attributes["center_y"].Value));
            else
                projector = new Projector(
                    int.Parse(node.Attributes["resolution_x"].Value),
                    int.Parse(node.Attributes["resolution_y"].Value),
                    Angle.FromDegrees(int.Parse(node.Attributes["horizontal_angle"].Value)));
            TryToSetPose(node, projector);
            return projector;
        }

        private ActiveDisplay LoadDisplay(XmlNode node)
        {
            var display = new ActiveDisplay(
                int.Parse(node.Attributes["width"].Value),
                int.Parse(node.Attributes["resolution_x"].Value),
                int.Parse(node.Attributes["resolution_y"].Value));
            TryToSetPose(node, display);
            return display;
        }

        private WallElement LoadWall(XmlNode node)
        {
            return new WallElement(
                LoadPoint(node["position"]),
                LoadVector(node["normal"]),
                LoadVector(node["up"]));
        }
        #endregion

        #region Load Primitives
        private Point3D LoadPoint(XmlNode node)
        {
            return new Point3D(
                double.Parse(node.Attributes["x"].Value),
                double.Parse(node.Attributes["y"].Value),
                double.Parse(node.Attributes["z"].Value));
        }

        private Vector3D LoadVector(XmlNode node)
        {
            return new Vector3D(
                double.Parse(node.Attributes["x"].Value),
                double.Parse(node.Attributes["y"].Value),
                double.Parse(node.Attributes["z"].Value));
        }

        private void TryToSetPose(XmlNode node, AbstractElement element)
        {
            if (node["position"] != null)
                element.SetPosition(LoadPoint(node["position"]));

            if (node["direction"] != null && node["up"] != null)
                element.SetOrientation(LoadVector(node["direction"]), LoadVector(node["up"]));
        }
        #endregion
    }
}
