﻿using System.Collections.Generic;
using ProjectorKit.Elements;
using ProjectorKit.World;

namespace ProjectorKit.Controller
{
    /// <summary>
    /// A Handler interacts with the world according to the usage of a button.
    /// Specific Handlers can be implemented by overriding HandleMove,
    /// HandlePressed and HandleReleased.
    /// 
    /// IMPORTANT: A world needs to be initialized before creating a Handler.
    /// </summary>
    public abstract class AbstractHandler
    {
        private static Dictionary<AbstractElement, AbstractElement> activeHandlers = new Dictionary<AbstractElement, AbstractElement>();
        private Button button;

        protected virtual void HandleMove(AbstractElement controller, AbstractElement target) { }
        protected virtual void HandlePressed(AbstractElement controller, AbstractElement target) { }
        protected virtual void HandleReleased(AbstractElement controller, AbstractElement target) { }

        public AbstractHandler(Button button)
        {
            this.button = button;
            button.Changed += HandleButton;
        }

        /// <summary>
        /// Restrict the elements handled by the handler. This method is called
        /// for the current target, everytime the handler is activated.
        /// </summary>
        /// <param name="element">Current target.</param>
        /// <returns>True if the handler can handle the element, otherwise false.</returns>
        protected abstract bool CanHandleElement(AbstractElement element);

        #region Internal
        private void HandleButton(AbstractElement sender, ButtonState state)
        {
            if (state == ButtonState.Pressed)
            {
                // Search closest element that handler can handle
                var targets = ServerWorld.Instance.PointingTargets(sender);
                for (int i = 0; i < targets.Count; i++)
                {
                    if (CanHandleElement(targets[i].Value))
                    {
                        HandleButtonPressedOn(sender, targets[i].Value);
                        break;
                    }
                }
            }
            else if (activeHandlers.ContainsKey(sender))
                HandleButtonPressedOff(sender);
        }

        private void HandleButtonPressedOn(AbstractElement sender, AbstractElement target)
        {
            if (CanHandleElement(target) && !activeHandlers.ContainsKey(sender))
            {
                activeHandlers.Add(sender, target);
                sender.Updated += ButtonMoved;
                HandlePressed(sender, target);
                ServerWorld.Instance.Focus(target);
                CursorDB.Instance.SetCursor(this.GetType(), sender);
            }
        }

        private void HandleButtonPressedOff(AbstractElement sender)
        {
            sender.Updated -= ButtonMoved;
            HandleReleased(sender, activeHandlers[sender]);
            CursorDB.Instance.SetDefaultCursor(sender);
            activeHandlers.Remove(sender);
        }

        private void ButtonMoved(AbstractElement element)
        {
            if (activeHandlers.ContainsKey(element))
                HandleMove((AbstractElement)element, activeHandlers[element]);
        }
        #endregion
    }
}
