﻿using System;
using System.Windows.Media.Media3D;
using GroupLab.iNetwork;
using ProjectorKit.Filter;
using ProjectorKit.Utilities;

namespace ProjectorKit.Elements
{
    /// <summary>
    /// The shared implementation of all element.
    /// </summary>
    public abstract class AbstractElement : IPositionable
    {
        private IFilter filter = new NoFilter();
        private Point3D position = new Point3D(0, 0, 0);
        private Vector3D up = new Vector3D(0, 1, 0);
        private Vector3D direction = new Vector3D(0, 0, 1);
        public delegate void UpdateEventHandler(AbstractElement element);
        public event UpdateEventHandler Updated = delegate { };

        #region Properties
        public virtual Point3D Position
        {
            get { return position; }
        }

        public virtual Vector3D Up
        {
            get { return up; }
        }

        public virtual Vector3D Direction
        {
            get { return direction; }
        }

        public Vector3D Right
        {
            get { return Vector3D.CrossProduct(direction, up); }
        }

        /// <summary>
        /// Set the filter of the element. This filter is applied in the 
        /// SetPose() function. Use a NoFilter object to avoid using a filter.
        /// </summary>
        public IFilter Filter
        {
            get { return filter; }
            set { filter = value; }
        }
        #endregion

        #region Pose setters
        /// <summary>
        /// Set the position of the element.
        /// </summary>
        /// <param name="position">New position</param>
        public void SetPosition(Point3D position)
        {
            SetPose(position, direction, up);
        }
        public void SetPosition(int x, int y, int z)
        {
            SetPosition(new Point3D(x, y, z));
        }

        /// <summary>
        /// Set the orientation of the element.
        /// </summary>
        /// <param name="direction">New direction vector</param>
        /// <param name="up">New up vector</param>
        public void SetOrientation(Vector3D direction, Vector3D up)
        {
            SetPose(position, direction, up);
        }

        /// <summary>
        /// Rotate element around a given axis.
        /// </summary>
        /// <param name="axis">Axis to use for rotation</param>
        /// <param name="angle">Angle to rotate</param>
        public void Rotate(Vector3D axis, Angle angle)
        {
            SetOrientation(
                Vector3DExtensions.RotationAroundArbitraryAxis(direction, Position, axis, angle),
                Vector3DExtensions.RotationAroundArbitraryAxis(up, Position, axis, angle));
        }

        /// <summary>
        /// Set the pose of the element. The final pose might be filtered.
        /// </summary>
        /// <param name="position">New position</param>
        /// <param name="direction">New direction vector</param>
        /// <param name="up">New up vector</param>
        public virtual void SetPose(Point3D position, Vector3D direction, Vector3D up)
        {
            // Replace parameters with filtered values
            filter.Filter(position, direction, up);
            position = filter.Position;
            direction = filter.Direction;
            up = filter.Up;

            direction.Normalize();
            up.Normalize();

            this.position = position;
            this.direction = direction;
            this.up = up;

            UpdateNotification();
        }
        #endregion

        #region WPF Mesh
        /// <summary>
        /// Mesh of the element. If it has no visual representation this is set to null.
        /// </summary>
        public virtual Model3D Mesh { get { return null; } }

        /// <summary>
        /// Updates the mesh.
        /// </summary>
        public virtual void UpdateMesh() { Model3D create = Mesh; }
        #endregion

        /// <summary>
        /// The type of the element in remote worlds. Override this if
        /// not all functionality is needed. Make sure ToMessage() and
        /// UpdateElement() work with the returned type.
        /// </summary>
        protected virtual Type NetworkType
        {
            get { return GetType(); }
        }

        /// <summary>
        /// Intersection of the given ray at the element.
        /// </summary>
        /// <param name="origin">Origin of the ray</param>
        /// <param name="direction">Direction of the ray</param>
        /// <returns>Intersection point or null if no intersection happens.</returns>
        public virtual Point3D? Intersection(Point3D origin, Vector3D direction)
        {
            return null;
        }

        protected void UpdateNotification()
        {
            Updated(this);
        }

        #region Synchronize
        /// <summary>
        /// Creates a message with the given name of the object.
        /// </summary>
        /// <param name="messageName">Name of the message</param>
        /// <returns>The message with object information</returns>
        public virtual Message ToMessage(string messageName)
        {
            Message msg = new Message(messageName);
            msg.Add("position", Position);
            msg.Add("up", Up);
            msg.Add("dir", Direction);
            msg.AddField("type", NetworkType.ToString());
            return msg;
        }

        /// <summary>
        /// Update the element with the given message. This method should call
        /// the update callback only once. Therefore always set attribute
        /// directly and call base.UpdateElement() last!
        /// </summary>
        /// <param name="msg">Message with update information</param>
        public virtual void UpdateElement(Message msg)
        {
            position = msg.GetPoint3D("position");
            direction = msg.GetVector3D("dir");
            up = msg.GetVector3D("up");
            UpdateNotification();
        }

        /// <summary>
        /// Create a new element from the given message.
        /// </summary>
        /// <param name="msg">Message with element information</param>
        /// <param name="modelGroup">The model group used by the element.</param>
        /// <returns>The new element</returns>
        public static AbstractElement CreateElement(Message msg, Model3DGroup modelGroup)
        {
            string sType = msg.GetStringField("type");
            Type type = Type.GetType(sType);
            AbstractElement element = (AbstractElement)Activator.CreateInstance(type);
            element.UpdateElement(msg);
            return element;
        }
        #endregion

        /// <summary>
        /// Calculate the intersection of the pointer to the object.
        /// If the pointer does not intersect the object, return null.
        /// </summary>
        /// <param name="pointer">Pointer of distance calculation</param>
        /// <returns>Distance from pointers position to the target</returns>
        public Point3D? Intersection(IPositionable pointer)
        {
            return Intersection(pointer.Position, pointer.Direction);
        }
    }
}
