﻿using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Media.Media3D;
using GroupLab.iNetwork;
using ProjectorKit.Utilities;
using System.Linq;

namespace ProjectorKit.Elements
{
    /// <summary>
    /// Viewports can be used for cameras in the Viewport3D.
    /// </summary>
    public abstract class AbstractDisplay : AbstractElement
    {
        private BlockingCollection<string> hiddenElements = new BlockingCollection<string>();
        private ConcurrentDictionary<string, double> blackouts = new ConcurrentDictionary<string, double>();
        private int resolutionX;
        private int resolutionY;
        #region Properties
        public int ResolutionX
        {
            get { return resolutionX; }
        }
        public int ResolutionY
        {
            get { return resolutionY; }
        }
        public int TotalPixels
        {
            get { return ResolutionX * ResolutionY; }
        }
        public double AspectRatio
        {
            get { return ResolutionX * 1.0 / ResolutionY; }
        }
        public Polygon PixelArea
        {
            get
            {
                return new Polygon(new Point[] { 
                    new Point(0,0), 
                    new Point(0, ResolutionY - 1), 
                    new Point(ResolutionX - 1, ResolutionY - 1), 
                    new Point(ResolutionX - 1, 0)});
            }
        }
        #endregion

        public Point3D PixelOrigin(Point pixel)
        {
            return PixelOrigin((int)pixel.X, (int)pixel.Y);
        }
        public Vector3D PixelDirection(Point pixel)
        {
            return PixelDirection((int)pixel.X, (int)pixel.Y);
        }

        /// <summary>
        /// The pixel origin of the display for an arbitrary pixel.
        /// The requested pixel might not really exist on the display.
        /// </summary>
        /// <param name="x">X coordinate of the pixel.</param>
        /// <param name="y">Y coordinate of the pixel.</param>
        /// <returns>The origin of given pixel.</returns>
        public abstract Point3D PixelOrigin(int x, int y);
        /// <summary>
        /// The pixel direction of the display for an arbitrary pixel.
        /// The requested pixel might not really exist on the display.
        /// </summary>
        /// <param name="x">X coordinate of the pixel.</param>
        /// <param name="y">Y coordinate of the pixel.</param>
        /// <returns>The direction of given pixel.</returns>
        public abstract Vector3D PixelDirection(int x, int y);
        public abstract Point ConvertToPixel(Point3D point);
        public abstract Camera Camera { get; }

        public AbstractDisplay() { }
        public AbstractDisplay(int resolutionX, int resolutionY)
        {
            this.resolutionX = resolutionX;
            this.resolutionY = resolutionY;
        }

        public bool PixelExists(int x, int y)
        {
            return x >= 0 && x < resolutionX && y >= 0 && y < resolutionY;
        }
        public bool PixelExists(Point pixel)
        {
            return PixelExists((int)pixel.X, (int)pixel.Y);
        }

        #region Blackouts/Clipping and Hidden
        public IDictionary<string, double> Blackouts
        {
            get { return blackouts; }
        }
        /// <summary>
        /// Instead of displaying the element, it only shows its shape with the the given opacity.
        /// </summary>
        /// <param name="elementName">Name of the element</param>
        /// <param name="opacity">Opacity ranging from 1 (black) to 0 (transparent)</param>
        public void AddBlackout(string elementName, double opacity = 1)
        {
            if (!blackouts.ContainsKey(elementName) || blackouts[elementName] != opacity)
            {
                blackouts[elementName] = opacity;
                UpdateNotification();
            }
        }
        public void RemoveBlackout(string elementName)
        {
            if (blackouts.ContainsKey(elementName))
            {
                double opacity;
                blackouts.TryRemove(elementName, out opacity);
                UpdateNotification();
            }
        }

        public IEnumerable<string> HiddenElements
        {
            get { return hiddenElements; }
        }
        public void HideElement(string elementName)
        {
            if (!hiddenElements.Contains(elementName) && hiddenElements.TryAdd(elementName))
                UpdateNotification();
        }
        public void ShowElement(string elementName)
        {
            if (hiddenElements.Contains(elementName) && hiddenElements.TryTake(out elementName))
                UpdateNotification();
        }
        #endregion

        #region Conversion to Pixels
        /// <summary>
        /// Converts an array of points with ConvertToPixels to coordinates.
        /// If one conversion is out of bounds the result is null.
        /// </summary>
        /// <param name="points">Points to convert</param>
        /// <returns>An array of coordinate for the given points.</returns>
        public Point[] ConvertToPixels(Point3D[] points)
        {
            Point[] result = new Point[points.Length];
            for (int i = 0; i < points.Length; i++)
            {
                Point? pixel = ConvertToPixel(points[i]);
                if (pixel != null)
                    result[i] = (Point)pixel;
                else
                    return null;
            }
            return result;
        }
        #endregion

        #region Synchronize
        public override void UpdateElement(Message msg)
        {
            resolutionX = msg.GetIntField("resolution_x");
            resolutionY = msg.GetIntField("resolution_y");
            blackouts = msg.GetConcurrentDictionaryStringDouble("blackout");
            hiddenElements = msg.GetBlockingCollectionString("hiddenElements");

            base.UpdateElement(msg);
        }

        public override Message ToMessage(string messageName)
        {
            Message msg = base.ToMessage(messageName);
            msg.AddField("resolution_x", resolutionX);
            msg.AddField("resolution_y", resolutionY);
            msg.Add("blackout", blackouts);
            msg.Add("hiddenElements", hiddenElements);

            return msg;
        }
        #endregion
    }
}
