﻿using System;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Media3D;
using GroupLab.iNetwork;
using ProjectorKit.Utilities;
using ProjectorKit.World;

namespace ProjectorKit.Elements
{
    public abstract class AbstractElement2D : RelativeElement, IPositionable2D
    {
        // The projector that projects the element; 'null' for augmentation.
        private string projectDevice = null;
        private ServerWorld world = null;
        private Point[] projectPixels;

        private GeometryModel3D mesh;
        private MeshGeometry3D geometry;
        private Polygon corners = new Polygon(new Point[0]);
        private bool blackout = false;
        #region Properties
        /// <summary>
        /// The projector name if the element is projected. If it is augmented: null.
        /// </summary>
        public string ProjectDevice
        {
            get { return projectDevice; }
        }
        /// <summary>
        /// Edges of the 2D element. These are the edges around the position of the element.
        /// Position, up and direction will change those coordinates.
        /// </summary>
        public virtual Polygon Corners
        {
            get { return corners; }
            set
            {
                if (value == null)
                    corners = new Polygon(new Point[0]);
                else if (value.Length < 3)
                    throw new ArgumentException("A 2D element needs at least 3 edges!");
                else
                    corners = value;
                UpdateNotification();
            }
        }

        public override Model3D Mesh
        {
            get
            {
                if (mesh == null)
                    CreateMesh();
                return mesh;
            }
        }

        /// <summary>
        /// Blacks out the element. The element is not getting transparent,
        /// all underlying elements are invisible.
        /// </summary>
        public bool Blackout
        {
            get { return blackout; }
            set
            {
                if (value != blackout)
                {
                    blackout = value;
                    UpdateNotification();
                }
            }
        }
        #endregion

        #region Constructor
        public AbstractElement2D() : base() { }
        public AbstractElement2D(double width, double height)
            : base()
        {
            SetRectangle(width, height);
        }
        public AbstractElement2D(Point3D position, double width, double height)
            : this(width, height)
        {
            SetPosition(position);
        }

        private void CreateMesh()
        {
            mesh = new GeometryModel3D();
            geometry = new MeshGeometry3D();

            PointCollection texture = new PointCollection();
            texture.Add(new Point(1, 0));
            texture.Add(new Point(1, 1));
            texture.Add(new Point(0, 1));
            texture.Add(new Point(0, 0));
            geometry.TextureCoordinates = texture;

            mesh.Material = new DiffuseMaterial(new SolidColorBrush(Colors.Red));
            mesh.Geometry = geometry;
        }
        #endregion

        #region ProjectRequirement
        /// <summary>
        /// Directly project it using the projector. Avoids jitter and keystone effects.
        /// Do not use this method directly. Use world.ProjectElement() instead!
        /// </summary>
        /// <param name="projector">Projector to use</param>
        /// <param name="projectorName">Name of the projector</param>
        /// <param name="x">X coordinate</param>
        /// <param name="y">Y coordinate</param>
        /// <param name="width">Width of the element</param>
        /// <param name="height">Height of the element</param>
        public void ProjectWith(ServerWorld world, Projector projector, string projectorName, int x, int y, int width, int height)
        {
            if(!projector.PixelExists(x, y) || !projector.PixelExists(x+width, y+height))
                throw new ArgumentException("The specified position and height/width does not fit into the projector.");

            this.world = world;
            projectDevice = projectorName;
            projectPixels = new Point[] { new Point(x, y), new Point(x+width, y), new Point(x+width, y+height), new Point(x, y+height) };
            projector.Updated += projectUpdate;
            refreshProject(projector);
        }

        private void refreshProject(Projector projector)
        {
            AbstractElement target = world.PointingTarget(projector.Position, projector.PixelDirection(projectPixels[0]));
            if (target != null)
            {
                // Get intersections
                Point3D?[] intersections = new Point3D?[] {
                    target.Intersection(projector.PixelOrigin(projectPixels[0]), projector.PixelDirection(projectPixels[0])),
                    target.Intersection(projector.PixelOrigin(projectPixels[1]), projector.PixelDirection(projectPixels[1])),
                    target.Intersection(projector.PixelOrigin(projectPixels[2]), projector.PixelDirection(projectPixels[2])),
                    target.Intersection(projector.PixelOrigin(projectPixels[3]), projector.PixelDirection(projectPixels[3]))
                };
                // Default location if no interaction on same target
                foreach (Point3D? intersect in intersections)
                {
                    if (intersect == null)
                        return;
                }

                // Rectangularize projection
                Point3D pos = Average(intersections).ToPoint3D();
                Plane plane = new Plane(pos, target.Direction);
                Corners = new Polygon(new Point[] {
                    plane.RelativeCoordinateFor((Point3D)intersections[1]),
                    plane.RelativeCoordinateFor((Point3D)intersections[2]),
                    plane.RelativeCoordinateFor((Point3D)intersections[3]),
                    plane.RelativeCoordinateFor((Point3D)intersections[0])
                });
                SetPose(pos, new Vector3D(0,0,-1), new Vector3D(0, 1, 0));
            }
        }
        private Vector3D Average(Point3D?[] points)
        {
            Vector3D result = new Vector3D();
            foreach (Point3D p in points)
                result += p.ToVector3D();
            return result / points.Length;
        }

        private void projectUpdate(AbstractElement element)
        {
            Projector projector = (Projector)element;
            Polygon3D p3d = new Polygon3D(Position, Direction, Corners);

            if (!projector.InProjection(p3d))
                refreshProject(projector);
        }
        #endregion

        protected abstract Material Texture();

        public override void UpdateMesh()
        {
            base.UpdateMesh();
            Point3DCollection positions = new Point3DCollection(4);
            for (int i = 0; i < corners.Length; i++)
                positions.Add(Position + corners[i].X * Right + Corners[i].Y * Up);
            geometry.Positions = positions;

            Vector3DCollection normals = new Vector3DCollection(corners.Length);
            foreach (Point3D p in geometry.Positions)
                normals.Add(Direction);
            geometry.Normals = normals;

            Int32Collection indices = new Int32Collection();
            for (int i = 1; i + 1 < geometry.Positions.Count; i++)
            {
                indices.Add(0);
                indices.Add(i + 1);
                indices.Add(i);
            }
            indices.Add(geometry.Positions.Count - 2);
            indices.Add(geometry.Positions.Count - 1);
            indices.Add(0);
            geometry.TriangleIndices = indices;

            if (blackout)
            {
                DiffuseMaterial material = new DiffuseMaterial();
                material.Brush = new SolidColorBrush(Colors.Black);
                mesh.Material = material;
            }
            else
                mesh.Material = Texture();
        }

        public override Point3D? Intersection(Point3D origin, Vector3D direction)
        {
            return (new Polygon3D(Position, Direction, Corners)).Intersection(origin, direction);
        }

        public void SetRectangle(double width, double height)
        {
            double hwidth = width / 2;
            double hheight = height / 2;
            Corners = new Polygon(new Point[] {
                    new Point(-hwidth, +hheight),
                    new Point(-hwidth, -hheight),
                    new Point(+hwidth, -hheight),
                    new Point(+hwidth, +hheight)
                });
        }

        #region Synchronization
        public override Message ToMessage(string messageName)
        {
            Message msg = base.ToMessage(messageName);
            msg.AddField("project_device", projectDevice);

            msg.AddField("edge_count", corners.Length);
            for (int i = 0; i < corners.Length; i++)
            {
                msg.AddField("edge_x" + i, corners.Points[i].X);
                msg.AddField("edge_y" + i, corners.Points[i].Y);
            }

            return msg;
        }

        public override void UpdateElement(Message msg)
        {
            projectDevice = msg.GetStringField("project_device");

            int edgeCount = msg.GetIntField("edge_count");
            corners = new Polygon(new Point[edgeCount]);
            for (int i = 0; i < edgeCount; i++)
            {
                corners.Points[i] = new Point(
                    msg.GetDoubleField("edge_x" + i),
                    msg.GetDoubleField("edge_y" + i)
                );
            }
            base.UpdateElement(msg);
        }
        #endregion
    }
}
