﻿using System.Windows.Media.Media3D;
using ProjectorKit.Elements;
using ProjectorKit.Utilities;
using ProjectorKit.World;

namespace ProjectorKit.Events
{
    public class PointingEventArgs : InteractionEventArgs
    {
        public AbstractElement Pointer;
        public AbstractElement Target;
        public Point3D Intersection;
        public Vector3D IntersectionFromCenter;
        public Angle PointerRotation;
        public double Distance;
        public double PixelSize;
    }

    /// <summary>
    /// This relation triggers callbacks when two elements are pointing to each other.
    /// </summary>
    public sealed class PointingEvent : AbstractInteraction<PointingEventArgs>
    {
        public delegate void PointingRelationEventHandler(object sender, PointingEventArgs e);
        public event PointingRelationEventHandler Pointed = delegate { };
        public event PointingRelationEventHandler PointingStarted = delegate { };
        public event PointingRelationEventHandler PointingChanged = delegate { };
        public event PointingRelationEventHandler PointingEnded = delegate { };

        private AbstractElement pointer;
        private AbstractElement target;
        private bool isInUpdate = false;

        public PointingEvent(AbstractElement pointer, AbstractElement target)
        {
            this.pointer = pointer;
            this.target = target;

            pointer.Updated += NoRecursiveUpdate;
            target.Updated += NoRecursiveUpdate;
        }

        protected override void Update(AbstractElement element)
        {
            AbstractElement target = ServerWorld.Instance.PointingTarget(pointer);

            UpdateStateWithActivityFlag(this.target == target);
            if (State != InteractionState.None)
            {
                PointingEventArgs args = CreateEventArgs();

                // Call the callbacks
                if (State == InteractionState.Start)
                    PointingStarted(this, args);
                if (State == InteractionState.Move)
                    PointingChanged(this, args);
                if (State == InteractionState.End)
                    PointingEnded(this, args);
                Pointed(this, args);
            }
        }

        private PointingEventArgs CreateEventArgs()
        {
            PointingEventArgs args = new PointingEventArgs();
            SetDefaultEventArgs(args);
            args.Pointer = pointer;
            args.Target = target;

            Plane plane = new Plane(pointer.Position, pointer.Direction);
            args.PointerRotation = Angle.Between(pointer.Up, plane.Up);

            Point3D? hitpoint = target.Intersection(pointer);
            if (hitpoint != null)
            {
                args.Intersection = (Point3D)hitpoint;
                args.IntersectionFromCenter = (Point3D)hitpoint - target.Position;
                args.Distance = pointer.Position.DistanceTo(args.Intersection);
                args.PixelSize = 1;
            }
            return args;
        }
    }
}
