﻿using System.Collections.Generic;
using System.Windows.Media.Media3D;
using ProjectorKit.Elements;
using ProjectorKit.Utilities;

namespace ProjectorKit.Events
{
    public class SelectionEventArgs : OverlappingEventArgs
    {
        public AbstractDisplay Projector;
        public ProjectorArgs ProjectorArgs;
        public AbstractElement2D Element;
        public ElementArgs ElementArgs;

        public SelectionEventArgs(Overlap overlapping) : base(overlapping) { }
    }

    /// <summary>
    /// This relation triggers callbacks when a display shows a hotspot.
    /// </summary>
    public sealed class SelectionEvent : AbstractOverlappingEvent<SelectionEventArgs>
    {
        public delegate void HotspotEventHandler(object sender, SelectionEventArgs e);
        public event HotspotEventHandler Overlapped = delegate { };
        public event HotspotEventHandler OverlappingStarted = delegate { };
        public event HotspotEventHandler OverlappingChanged = delegate { };
        public event HotspotEventHandler OverlappingEnded = delegate { };
        private AbstractDisplay projector;
        private AbstractElement2D element;
        private Polygon area;

        #region Constructors
        public SelectionEvent(AbstractElement2D element, AbstractDisplay projector)
            : this(element, projector.PixelArea, projector) { }

        /// <summary>
        /// Create an OverlappingRelation between a display and an element.
        /// If no area is specified, the whole viewport is used.
        /// </summary>
        /// <param name="element">The element of the overlapping</param>
        /// <param name="area">[Optional] Polygon of pixels specifying the area in the viewport 1</param>
        /// <param name="display">The display of the overlapping</param>
        public SelectionEvent(AbstractElement2D element, Polygon area, AbstractDisplay display)
        {
            this.projector = display;
            this.area = area;
            this.element = element;

            display.Updated += NoRecursiveUpdate;
            element.Updated += NoRecursiveUpdate;
        }
        #endregion

        protected override void Update(AbstractElement element)
        {
            Overlap overlapping = Overlap.Of(projector, this.element, area);

            UpdateStateWithActivityFlag(overlapping != null);
            SelectionEventArgs args = CreateEventArgs(overlapping);
            if (State != InteractionState.None)
            {
                if (State == InteractionState.Start)
                    OverlappingStarted(this, args);
                if (State == InteractionState.Move)
                    OverlappingChanged(this, args);
                if (State == InteractionState.End)
                    OverlappingEnded(this, args);
                Overlapped(this, args);
            }
        }

        private SelectionEventArgs CreateEventArgs(Overlap overlapping)
        {
            SelectionEventArgs args = new SelectionEventArgs(overlapping);
            SetDefaultEventArgs(args);
            args.Projector = projector;
            args.Element = element;
            if (overlapping != null)
            {
                args.ProjectorArgs = new ProjectorArgs(projector, area, overlapping);
                args.ElementArgs = new ElementArgs(element, overlapping);
            }
            return args;
        }
    }
}
