﻿using System.Windows;
using System.Windows.Media.Media3D;
using ProjectorKit.Elements;
using ProjectorKit.Utilities;
using ProjectorKit.World;

namespace ProjectorKit.Events
{
    public class VisibilityEventArgs : InteractionEventArgs
    {
        public AbstractElement Pointer;
        public Point3D VisiblePoint;
        public Point Pixel;
        public AbstractElement HidingElement;
    }

    /// <summary>
    /// A relation that triggers a callback when a given point is visible.
    /// </summary>
    public sealed class VisiblePointEvent : AbstractInteraction<VisibilityEventArgs>
    {
        public delegate void VisibilityRelationEventHandler(object sender, VisibilityEventArgs e);
        public event VisibilityRelationEventHandler Visible = delegate { };
        public event VisibilityRelationEventHandler VisibleStart = delegate { };
        public event VisibilityRelationEventHandler VisibleChanged = delegate { };
        public event VisibilityRelationEventHandler VisibleEnd = delegate { };

        private AbstractDisplay viewport;
        private Point3D point;

        public VisiblePointEvent(AbstractDisplay viewport, Point3D point)
        {
            this.viewport = viewport;
            this.point = point;

            viewport.Updated += NoRecursiveUpdate;
        }

        /// <summary>
        /// Changes the point of the relation.
        /// </summary>
        /// <param name="point">New point of the relation</param>
        public void ChangePoint(Point3D point)
        {
            this.point = point;
            Update(viewport);
        }

        protected override void Update(AbstractElement element)
        {
            Point pixel = viewport.ConvertToPixel(point);
            bool visible = viewport.PixelExists((int)pixel.X, (int)pixel.Y);
            UpdateStateWithActivityFlag(visible);
            if(State != InteractionState.None)
            {
                double distance = point.DistanceTo(viewport.Position);
                AbstractElement target = ServerWorld.Instance.PointingTarget(
                    viewport.PixelOrigin((int)pixel.X, (int)pixel.Y),
                    viewport.PixelDirection((int)pixel.X, (int)pixel.Y));
                if (target != null)
                {
                    Point3D? hit = target.Intersection(viewport);
                    if (hit != null && distance - ((Point3D)hit).DistanceTo(viewport.Position) < double.Epsilon)
                        target = null;
                }

                VisibilityEventArgs args = CreateEventArgs(pixel, null);
                if (State == InteractionState.Start)
                    VisibleStart(this, args);
                if (State == InteractionState.Move)
                    VisibleChanged(this, args);
                if (State == InteractionState.End)
                    VisibleEnd(this, args);
                Visible(this, args);
            }
        }

        private VisibilityEventArgs CreateEventArgs(Point pixel, AbstractElement hidingElement)
        {
            VisibilityEventArgs args = new VisibilityEventArgs();
            SetDefaultEventArgs(args);
            args.Pointer = viewport;
            args.VisiblePoint = point;
            args.Pixel = pixel;
            args.HidingElement = hidingElement;
            return args;
        }
    }
}
