﻿using ProjectorKit.Elements;
using ProjectorKit.Events;
using ProjectorKit.Utilities;
using System;
using System.Windows.Media.Media3D;

namespace ProjectorToolkit.Gestures
{
    public class RotatingEventArgs : InteractionEventArgs
    {
        public Vector3D Axis;
        public Angle Rotation;
    }

    public class RotatingGesture : AbstractInteraction<RotatingEventArgs>
    {
        public delegate void RotationEventHandler(object sender, RotatingEventArgs e);
        public event RotationEventHandler RotationStarted = delegate { };
        public event RotationEventHandler RotationChanged = delegate { };
        public event RotationEventHandler RotationEnded = delegate { };
        public event RotationEventHandler Rotation = delegate { };

        private AbstractElement element;
        private Angle minRotation;
        private double degreePerSecond;
        private Vector3D oldVector;
        private Vector3D lastVector;
        private DateTime lastTime;

        /// <summary>
        /// Fires a rotation gesture.
        /// </summary>
        /// <param name="element">Element of the gesture.</param>
        /// <param name="minRotation">Minimum rotation before fired</param>
        /// <param name="degreePerSecond">Minimum degrees per second of the rotation</param>       

        public RotatingGesture(AbstractElement element, Angle minRotation, double degreePerSecond=1)
        {
            this.element = element;
            this.minRotation = minRotation;
            this.degreePerSecond = degreePerSecond;
            this.oldVector = element.Direction;
            this.lastVector = element.Direction;

            element.Updated += NoRecursiveUpdate;
        }

        protected override void Update(AbstractElement element)
        {
            bool active = true;
            DateTime now = DateTime.Now;

            Angle diff = Angle.Between(element.Direction, oldVector);
            // Rotation too small
            if (diff.InDegrees < minRotation.InDegrees)
                active = false;
            // Timeout when degrees per second are too low
            double degPerSec = (Angle.Between(lastVector, element.Direction)).InDegrees / (now - lastTime).TotalSeconds;
            if (degreePerSecond > degPerSec || double.IsInfinity(degPerSec))
            {
                active = false;
                if(!double.IsInfinity(degPerSec))
                    oldVector = element.Direction;
            }
            lastVector = element.Direction;
            lastTime = now;

            UpdateStateWithActivityFlag(active);
            if (State != InteractionState.None)
            {
                RotatingEventArgs args = CreateEventArgs();

                // Call the callbacks
                if (State == InteractionState.Start)
                    RotationStarted(this, args);
                if (State == InteractionState.Move)
                    RotationChanged(this, args);
                if (State == InteractionState.End)
                {
                    RotationEnded(this, args);
                    oldVector = element.Direction;
                }
                Rotation(this, args);
            }
        }

        private RotatingEventArgs CreateEventArgs()
        {
            RotatingEventArgs args = new RotatingEventArgs();
            SetDefaultEventArgs(args);

            args.Axis = Vector3D.CrossProduct(element.Direction, oldVector);
            args.Axis.Normalize();
            args.Rotation = Angle.Between(element.Direction, oldVector);
            return args;
        }
    }
}
