﻿using ProjectorKit.Utilities;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Windows.Media.Media3D;

namespace ProjectorToolkit.Utilities
{
    public class AnalyzedData
    {
        // Merge lines smaller than this angle
        private const double MERGE_ANGLE = 45;
        public LinkedList<Vector3D> MergedSamples = new LinkedList<Vector3D>();
        public DateTime Timestamp;
        public double Peak;
        public Vector3D MainAxis;

        public AnalyzedData(Point3D[] samples)
        {
            Timestamp = DateTime.Now;
            Peak = 0;
            if (samples.Length >= 3)
            {
                // Approximate a zig-zag line by merging samples
                bool hardCorner = true;
                Vector3D lastVector = (samples[1].ToVector3D() - samples[0].ToVector3D());
                Vector3D line = new Vector3D(0, 0, 0);
                for (int i = 2; i < samples.Length; i++)
                {
                    Vector3D newVector = (samples[i].ToVector3D() - samples[i - 1].ToVector3D());

                    // Merge as long as direction is similar
                    Angle corner = Angle.Between(lastVector, newVector);
                    if (hardCorner || corner.InDegrees <= MERGE_ANGLE)
                    {
                        line += newVector;
                        lastVector = newVector;
                        hardCorner = false;
                    }
                    else // This is a hard corner, do not merge
                    {
                        // Peak is the longest vector
                        if (line.Length >= Peak)
                            Peak = line.Length;
                        MergedSamples.AddLast(line);
                        // Reset line for next iteration
                        line = new Vector3D(0, 0, 0);
                        hardCorner = true;
                    }
                }
                // Check for last iteration
                if (!hardCorner)
                {
                    // Peak is the longest vector
                    if (line.Length >= Peak)
                        Peak = line.Length;
                    MergedSamples.AddLast(line);
                }

                // Calculate main axis
                MainAxis = new Vector3D(0, 0, 0);
                foreach (Vector3D sample in MergedSamples)
                {
                    Vector3D vector = sample.Duplicate();
                    if (Vector3DExtensions.SameDirection(vector, MainAxis))
                        vector.Negate();
                    MainAxis += vector;
                }
                MainAxis /= MergedSamples.Count;
            }
        }

        public double PeakFor(Vector3D axis, Angle angleTolerance)
        {
            double peak = 0;
            foreach (Vector3D sample in MergedSamples)
            {
                Vector3D vector = sample.Duplicate();
                if (Vector3DExtensions.SameDirection(vector, MainAxis))
                    vector.Negate();

                if (Angle.Between(axis, vector).InDegrees <= angleTolerance.InDegrees)
                {
                    if (vector.Length > peak)
                        peak = vector.Length;
                }
            }
            return peak;
        }

        public int CountFor(Vector3D axis, Angle angleTolerance)
        {
            int count = 0;
            foreach (Vector3D sample in MergedSamples)
            {
                Vector3D vector = sample.Duplicate();
                if (Vector3DExtensions.SameDirection(vector, MainAxis))
                    vector.Negate();

                if (Angle.Between(axis, vector).InDegrees <= angleTolerance.InDegrees)
                    count++;
            }
            return count;
        }
    }

    public class MotionAnalyzer
    {
        private double duration;
        private LinkedList<KeyValuePair<DateTime, Point3D>> samples = new LinkedList<KeyValuePair<DateTime, Point3D>>();

        /// <summary>
        /// Analyzes motions.
        /// </summary>
        /// <param name="duration">Maximum duration to listen to [in s].</param>
        public MotionAnalyzer(double duration)
        {
            this.duration = duration;
        }

        public void Reset()
        {
            samples.Clear();
        }

        public void AddPoint(Point3D point)
        {
            samples.AddLast(new KeyValuePair<DateTime, Point3D>(DateTime.Now, point));
            PruneOldSamples();
        }

        private void PruneOldSamples()
        {
            DateTime now = DateTime.Now;
            while (samples.First != null && (now - samples.First.Value.Key).Seconds > duration)
                samples.RemoveFirst();
        }

        public AnalyzedData Analyze()
        {
            Point3D[] points = new Point3D[samples.Count];
            var node = samples.First;
            for(int i = 0; i<samples.Count; i++)
            {
                points[i] = node.Value.Value;
                node = node.Next;
            }
            return new AnalyzedData(points);
        }
    }
}
