﻿
namespace ProjectorKit.Utilities
{
    /// <summary>
    /// Object pool to recycle old objects to increase performance.
    /// Allocate does not reset old properties!
    /// </summary>
    /// <typeparam name="T">Class type to heap.</typeparam>
    public sealed class ObjectPool<T> where T : class, new()
    {
        #region Stack Implementation
        /// <summary>
        /// Efficient heapstack implementation. All functions take O(1).
        /// </summary>
        /// <typeparam name="T">Type of the values.</typeparam>
        public class HeapStack<T>
        {
            private HeapNode First = null;
            private HeapNode Last = null;

            public class HeapNode
            {
                public T Value;
                public HeapNode Previous = null;

                public HeapNode(T value, HeapNode previous = null)
                {
                    this.Value = value;
                    this.Previous = previous;
                }
            }

            public bool IsEmpty()
            {
                return Last == null;
            }

            public void Push(HeapNode node)
            {
                if (IsEmpty())
                    First = node;
                Last = node;
            }
            public void Push(T value)
            {
                HeapNode node = new HeapNode(value, Last);
                Push(node);
            }

            public HeapNode Pop()
            {
                HeapNode result = Last;
                Last = Last.Previous;

                if (IsEmpty())
                    First = null;

                return result;
            }

            public void Append(HeapStack<T> stack)
            {
                if (!stack.IsEmpty())
                {
                    stack.First.Previous = Last;
                    Last = stack.Last;
                }
            }

            public void Clear()
            {
                First = null;
                Last = null;
            }
        }
        #endregion
        private readonly HeapStack<T> recycled = new HeapStack<T>();
        private readonly HeapStack<T> active = new HeapStack<T>();

        /// <summary>
        /// Allocates memory for a new object or reuses old one.
        /// </summary>
        /// <returns></returns>
        public T Allocate()
        {
            if (recycled.IsEmpty())
            {
                T result = new T();
                active.Push(result);
                return result;
            }
            else
            {
                HeapStack<T>.HeapNode result = recycled.Pop();
                active.Push(result);
                return result.Value;
            }
        }

        /// <summary>
        /// Puts all allocated objects into the recycle queue.
        /// </summary>
        public void RecycleAll()
        {
            recycled.Append(active);
            active.Clear();
        }

        /// <summary>
        /// Unloads all allocated active and recycled objects. The objects 
        /// cannot be reused and other references may hold them in memory.
        /// </summary>
        public void ReleaseAll()
        {
            recycled.Clear();
            active.Clear();
        }
    }
}
