﻿using System.Collections.Generic;
using System.Windows;
using System.Windows.Media.Media3D;
using ProjectorKit.Elements;

namespace ProjectorKit.Utilities
{
    public class Overlap
    {
        private Plane plane;
        private List<Polygon> overlaps;
        #region Properties
        public Plane Plane
        {
            get { return plane; }
        }
        public List<Polygon> Overlaps
        {
            get { return overlaps; }
        }
        #endregion

        #region Static Creation
        public static Overlap Of(AbstractDisplay display1, AbstractDisplay display2)
        {
            return Of(display1, display1.PixelArea, display2, display2.PixelArea);
        }
        public static Overlap Of(AbstractDisplay display1, Polygon area1, AbstractDisplay display2)
        {
            return Of(display1, area1, display2, display2.PixelArea);
        }
        public static Overlap Of(AbstractDisplay display1, Polygon area1, AbstractDisplay display2, Polygon area2)
        {
            Plane xyPlane = new Plane(new Point3D(0, 0, 0), new Vector3D(0, 0, 1));
            Polygon polygon1 = IntersectionWithPlane(xyPlane, display1, area1);
            Polygon polygon2 = IntersectionWithPlane(xyPlane, display2, area2);
            return CreateResult(xyPlane, polygon1, polygon2);
        }

        public static Overlap Of(AbstractDisplay display, IPositionable2D element)
        {
            return Of(display, display.PixelArea, element);
        }
        public static Overlap Of(AbstractDisplay display, Polygon area, IPositionable2D element)
        {
            Plane plane = new Plane(element.Position, element.Direction);
            Polygon polygon1 = IntersectionWithPlane(plane, display, area);
            Polygon polygon2 = element.Corners;
            return CreateResult(plane, polygon1, polygon2);
        }
        public static Overlap Of(AbstractDisplay display, IPositionable2D element, Polygon area)
        {
            Plane plane = new Plane(element.Position, element.Direction);
            Polygon polygon1 = IntersectionWithPlane(plane, display, element.Corners);
            Polygon polygon2 = area;
            return CreateResult(plane, polygon1, polygon2);
        }

        public static Overlap Of(IPositionable2D element1, IPositionable2D element2)
        {
            Plane plane1 = new Plane(element1.Position, element1.Direction);
            Plane plane2 = new Plane(element2.Position, element2.Direction);

            // Project corners of element 1 onto plane of element 2
            Polygon polygon1 = new Polygon(new Point[element1.Corners.Points.Length]);
            for (int i = 0; i < element1.Corners.Points.Length; i++)
            {
                Point3D origin = plane1.AbsoluteCoordinateFor(element1.Corners[i]);
                Point3D? hit = plane2.Intersection(origin, plane1.Normal);
                if (hit == null)
                    return null;

                polygon1.Points[i] = plane2.RelativeCoordinateFor((Point3D)hit);
            }

            return CreateResult(plane2, polygon1, element2.Corners);
        }

        /// <summary>
        /// Intersects the area of the display with a plane.
        /// </summary>
        /// <param name="plane">Plane to intersect</param>
        /// <param name="display">Display for intersection</param>
        /// <param name="area">Pixel area of display to intersect</param>
        /// <returns>Intersection on the plane or null, if no intersection was possible,</returns>
        private static Polygon IntersectionWithPlane(Plane plane, AbstractDisplay display, Polygon area)
        {
            List<Point> points = new List<Point>();
            foreach (Point p in area.Points)
            {
                Point3D? hit = plane.Intersection(display.PixelOrigin(p), display.PixelDirection(p));
                if (hit != null)
                    points.Add(plane.RelativeCoordinateFor((Point3D)hit));
                else
                    return null;
            }
            return NormalizePolygon(new Polygon(points.ToArray()), display.Direction, plane.Normal);
        }

        /// <summary>
        /// Makes a anti-clockwise polygon out of the given one.
        /// </summary>
        /// <param name="polygon">Polygon to normalize</param>
        /// <param name="direction">Direction of the creator</param>
        /// <param name="normal">Normal of the plane</param>
        /// <returns>Normalized polygon.</returns>
        private static Polygon NormalizePolygon(Polygon polygon, Vector3D direction, Vector3D normal)
        {
            if (Vector3D.DotProduct(direction, normal) < 0)
                return polygon.ChangeDirection();
            else
                return polygon;
        }

        private static Overlap CreateResult(Plane plane, Polygon polygon1, Polygon polygon2)
        {
            if (plane != null && polygon1 != null && polygon2 != null)
            {
                List<Polygon> overlaps = polygon1.IntersectWith(polygon2);
                if (overlaps != null && overlaps.Count > 0)
                    return new Overlap(plane, overlaps);
            }
            return null;
        }
        #endregion

        private Overlap(Plane plane, List<Polygon> overlappings)
        {
            this.plane = plane;
            this.overlaps = overlappings;
        }

        public double Area
        {
            get
            {
                double sum = 0;
                foreach (Polygon polygon in Overlaps)
                    sum += polygon.Area;
                return sum;
            }
        }

        public Point3D[] PolygonToAbsolute(Polygon polygon)
        {
            Point3D[] points = new Point3D[polygon.Points.Length];
            for (int i = 0; i < polygon.Length; i++)
                points[i] = Plane.AbsoluteCoordinateFor(polygon.Points[i]);
            return points;
        }
    }
}
