﻿using System;
using System.Windows;
using System.Windows.Media.Media3D;
using ProjectorKit.Elements;

namespace ProjectorKit.Utilities
{
    public class Plane
    {
        private Point3D point;
        private Vector3D normal;

        #region Properties
        public Point3D Point
        {
            get { return point; }
        }

        public Vector3D Normal
        {
            get { return normal; }
        }

        public Vector3D Up
        {
            get
            {
                Vector3D absoluteUp = new Vector3D(0, 1, 0);

                // Trivial 1: Plane is orthogonal to absolute up -> Choose arbitrary
                if (normal == absoluteUp || normal == new Vector3D(0, -1, 0))
                    return new Vector3D(0, 0, 1);

                // Trivial 2: Plane shares the absolute up vector -> Use it
                if (Vector3D.DotProduct(normal, absoluteUp) == 0)
                    return absoluteUp;

                // Use vector on plane that is most similar to the absolute up vector.
                //               h
                //              /|
                //             / | (^ absoluteUp)
                //            /  | 
                //           /   |
                // (point)  p----o---- (> normal)
                //         / 
                // Calculate a point o outside of the plane that differs from p
                // only in the direction of normal n. Intersect plane from o with
                // vector the absolute up vector -> h. Most similar vector is (h-p).
                Point3D offPlanePoint = (point + normal);
                Point3D hit = (Point3D)Intersection(offPlanePoint, absoluteUp);

                // Calculate the hit vector
                Vector3D up = (point - hit);
                up.Normalize();

                // Swap if the vector is to the wrong direction
                if (up.Y < 0)
                    up.Negate();

                return up;
            }
        }

        public Vector3D Right
        {
            get { return Vector3D.CrossProduct(normal, Up); }
        }
        #endregion

        public Plane(Point3D point, Vector3D normal)
        {
            this.point = point;
            this.normal = normal;
        }

        public Point3D? Intersection(IPositionable pointer)
        {
            return Intersection(pointer.Position, pointer.Direction);
        }

        public Point3D? Intersection(Point3D origin, Vector3D direction)
        {
            double distance = Distance(origin, direction);
            if (double.IsNaN(distance))
                return null;
            else
                return origin + distance * direction;
        }

        public double Distance(Point3D point)
        {
            return Math.Abs(Distance(point, normal));
        }

        public double Distance(Point3D origin, Vector3D direction)
        {
            double divisor = Vector3D.DotProduct(direction, normal);
            if (divisor != 0)
                return Vector3D.DotProduct((point - origin), normal) / divisor;
            else
                return double.NaN;
        }

        public Point RelativeCoordinateFor(Point3D point)
        {
            return new Point(
                Vector3D.DotProduct(point - this.point, Right),
                Vector3D.DotProduct(point - this.point, Up));
        }

        public Point3D AbsoluteCoordinateFor(Point point)
        {
            return this.point + point.X * Right + point.Y * Up;
        }
    }
}
