﻿using System;
using System.Collections.Generic;
using System.Windows;

namespace ProjectorKit.Utilities
{
    /// <author>Sebastian Boring</author>
    internal class LineSegment
    {
        #region Class Members
        private Point _start;

        private Point _end;
        #endregion

        #region Constructors
        internal LineSegment(Point start, Point end)
        {
            _start = start;
            _end = end;
        }
        #endregion

        #region Properties
        internal Point Start
        {
            get { return _start; }
        }

        internal Point End
        {
            get { return _end; }
        }
        #endregion

        #region Methods
        internal bool IntersectWith(LineSegment segment, ref Point intersection)
        {
            double q = (Start.Y - segment.Start.Y) * (segment.End.X - segment.Start.X)
                - (Start.X - segment.Start.X) * (segment.End.Y - segment.Start.Y);
            double d = (End.X - Start.X) * (segment.End.Y - segment.Start.Y)
                - (End.Y - Start.Y) * (segment.End.X - segment.Start.X);

            if (Math.Abs(d) <= 0.0001)
            {
                return false;
            }

            double r = q / d;
            q = (Start.Y - segment.Start.Y) * (End.X - Start.X)
                - (Start.X - segment.Start.X) * (End.Y - Start.Y);
            double s = q / d;

            if (r < 0 || r > 1 || s < 0 || s > 1)
            {
                return false;
            }

            intersection.X = Start.X + r * (End.X - Start.X);
            intersection.Y = Start.Y + r * (End.Y - Start.Y);

            return true;
        }
        #endregion
    }

    /// <author>Sebastian Boring</author>
    public class Polygon
    {
        #region Nested Class 'PolygonClipper'
        internal class PolygonClipper
        {
            #region Intersection Methods
            private static PolygonPoint FindEnteringIntersection(List<PolygonPoint> subjectPoints)
            {
                for (int i = 0; i < subjectPoints.Count; i++)
                {
                    if (!(subjectPoints[i].IsIntersection)
                        || subjectPoints[i].IsVisited)
                    {
                        continue;
                    }

                    if (subjectPoints[i].IsInbound)
                    {
                        return (subjectPoints[i]);
                    }
                }

                return null;
            }

            private static bool IntersectLines(Point start1, Point end1,
                Point start2, Point end2, ref Point intersection)
            {
                double q = (start1.Y - start2.Y) * (end2.X - start2.X)
                    - (start1.X - start2.X) * (end2.Y - start2.Y);
                double d = (end1.X - start1.X) * (end2.Y - start2.Y)
                    - (end1.Y - start1.Y) * (end2.X - start2.X);

                if (Math.Abs(d) <= 0.0001f)
                {
                    return false;
                }

                double r = q / d;
                q = (start1.Y - start2.Y) * (end1.X - start1.X)
                    - (start1.X - start2.X) * (end1.Y - start1.Y);
                double s = q / d;

                if (r < 0 || r > 1 || s < 0 || s > 1)
                {
                    return false;
                }

                intersection.X = start1.X + r * (end1.X - start1.X);
                intersection.Y = start1.Y + r * (end1.Y - start1.Y);

                return true;
            }

            private static void InsertIntersection(List<PolygonPoint> list,
                PolygonPoint intersectionPt, int start, int end)
            {
                double distance = Math.Sqrt(Math.Pow(list[start].Point.X - intersectionPt.Point.X, 2)
                    + Math.Pow(list[start].Point.Y - intersectionPt.Point.Y, 2));
                int insertIndex = -1;

                if (end > start)
                {
                    insertIndex = end;
                    for (int i = start + 1; i < end; i++)
                    {
                        double currDistance = Math.Sqrt(Math.Pow(list[start].Point.X - list[i].Point.X, 2)
                            + Math.Pow(list[start].Point.Y - list[i].Point.Y, 2));
                        if (currDistance > distance)
                        {
                            insertIndex = i;
                            break;
                        }
                    }
                }
                else
                {
                    insertIndex = start + 1;
                    bool added = false;
                    for (int i = start + 1; i < list.Count; i++)
                    {
                        double currDistance = Math.Sqrt(Math.Pow(list[start].Point.X - list[i].Point.X, 2)
                            + Math.Pow(list[start].Point.Y - list[i].Point.Y, 2));
                        if (currDistance > distance)
                        {
                            insertIndex = i;
                            added = true;
                            break;
                        }
                    }

                    if (!(added))
                    {
                        insertIndex = list.Count;
                    }
                }

                // now we insert the point
                if (insertIndex == list.Count)
                {
                    // add it to the end of the list
                    list[list.Count - 1].Next = intersectionPt;
                    intersectionPt.Next = list[0];

                    list.Add(intersectionPt);
                }
                else
                {
                    // insert the point
                    list[insertIndex - 1].Next = intersectionPt;
                    intersectionPt.Next = list[insertIndex];

                    list.Insert(insertIndex, intersectionPt);
                }
            }

            private static bool IsInbound(Point a, Point b, Point c, Point d)
            {
                Point vectB = new Point(b.X - a.X, b.Y - a.Y);
                Point vectD = new Point(d.X - c.X, d.Y - c.Y);

                vectD = new Point(-vectD.Y, vectD.X);
                double denominator = vectD.X * vectB.X + vectD.Y * vectB.Y;

                if (denominator > 0)
                {
                    return true;
                }

                return false;
            }

            private static void FindIntersections(List<PolygonPoint> subjectPolygon, List<PolygonPoint> clipPolygon)
            {
                int intersectionIndex = 0;
                for (int i = 0; i < subjectPolygon.Count; i++)
                {
                    if (subjectPolygon[i].IsIntersection)
                    {
                        continue;
                    }

                    PolygonPoint nextSubject = subjectPolygon[i].Next;
                    while (nextSubject.IsIntersection)
                    {
                        nextSubject = nextSubject.Next;
                    }

                    for (int j = 0; j < clipPolygon.Count; j++)
                    {
                        if (clipPolygon[j].IsIntersection)
                        {
                            continue;
                        }

                        PolygonPoint nextClip = clipPolygon[j].Next;
                        while (nextClip.IsIntersection)
                        {
                            nextClip = nextClip.Next;
                        }

                        Point intersection = new Point(0.0, 0.0);
                        bool success = IntersectLines(subjectPolygon[i].Point, nextSubject.Point,
                            clipPolygon[j].Point, nextClip.Point, ref intersection);

                        if (success)
                        {
                            // we need to insert the point
                            PolygonPoint clipIntersectionPt = new PolygonPoint(intersection);
                            PolygonPoint subjectIntersectionPt = new PolygonPoint(intersection);

                            clipIntersectionPt.Id = intersectionIndex;
                            subjectIntersectionPt.Id = intersectionIndex;

                            bool inbound = IsInbound(subjectPolygon[i].Point,
                                nextSubject.Point, clipPolygon[j].Point, nextClip.Point);

                            clipIntersectionPt.IsInbound = inbound;
                            subjectIntersectionPt.IsInbound = inbound;

                            intersectionIndex++;

                            clipIntersectionPt.IsIntersection = true;
                            subjectIntersectionPt.IsIntersection = true;

                            InsertIntersection(clipPolygon, clipIntersectionPt,
                                clipPolygon.IndexOf(clipPolygon[j]),
                                clipPolygon.IndexOf(nextClip));

                            InsertIntersection(subjectPolygon, subjectIntersectionPt,
                                subjectPolygon.IndexOf(subjectPolygon[i]),
                                subjectPolygon.IndexOf(nextSubject));
                        }
                    }
                }
            }
            #endregion

            #region Point Methods
            private static List<PolygonPoint> GetPolygonPoints(Polygon polygon)
            {
                List<PolygonPoint> polygonPts = new List<PolygonPoint>();
                int length = polygon.Points.Length;

                for (int i = 0; i < length; i++)
                {
                    PolygonPoint polygonPt = new PolygonPoint(polygon.Points[i]);
                    polygonPts.Add(polygonPt);

                    if (i != 0)
                    {
                        polygonPts[i - 1].Next = polygonPt;
                    }
                }
                polygonPts[length - 1].Next = polygonPts[0];

                return (polygonPts);
            }
            #endregion

            #region List Methods
            private static PolygonPoint GetIntersectionForId(int id, List<PolygonPoint> points)
            {
                for (int i = 0; i < points.Count; i++)
                {
                    if (points[i].Id == id)
                    {
                        return points[i];
                    }
                }
                return null;
            }
            #endregion

            #region Main Methods
            internal static List<Polygon> Clip(Polygon clipPolygon, Polygon subjectPolygon)
            {
                List<Polygon> polygons = new List<Polygon>();

                List<PolygonPoint> clipPoints = GetPolygonPoints(clipPolygon);
                List<PolygonPoint> subjectPoints = GetPolygonPoints(subjectPolygon);

                FindIntersections(subjectPoints, clipPoints);

                // run the processing
                // start with the subject
                PolygonPoint currentPoint = FindEnteringIntersection(subjectPoints);
                while (currentPoint != null)
                {
                    List<PolygonPoint> newPolygon = new List<PolygonPoint>();

                    currentPoint.IsVisited = true;
                    GetIntersectionForId(currentPoint.Id, clipPoints).IsVisited = true;

                    // add it to the list
                    newPolygon.Add(currentPoint);

                    List<PolygonPoint> currentList = subjectPoints;

                    PolygonPoint nextPt = currentPoint.Next;
                    while (!(nextPt.IsVisited))
                    {
                        newPolygon.Add(nextPt);
                        if (nextPt.IsIntersection)
                        {
                            nextPt.IsVisited = true;

                            // we switch the list
                            if (currentList.Equals(subjectPoints))
                            {
                                currentList = clipPoints;
                                nextPt = GetIntersectionForId(nextPt.Id, clipPoints);
                            }
                            else
                            {
                                currentList = subjectPoints;
                                nextPt = GetIntersectionForId(nextPt.Id, subjectPoints);
                            }
                            nextPt.IsVisited = true;
                            nextPt = nextPt.Next;
                        }
                        else
                        {
                            nextPt = nextPt.Next;
                        }
                    }

                    // ok, add the polygon to the final list
                    Point[] points = new Point[newPolygon.Count];
                    for (int i = 0; i < newPolygon.Count; i++)
                    {
                        points[i] = newPolygon[i].Point;
                    }
                    polygons.Add(new Polygon(points));

                    currentPoint = FindEnteringIntersection(subjectPoints);
                }

                if (polygons.Count == 0)
                {
                    bool inside = true;
                    for (int i = 0; i < subjectPolygon.Points.Length; i++)
                    {
                        if (!(clipPolygon.Contains(subjectPolygon[i])))
                        {
                            inside = false;
                            break;
                        }
                    }

                    if (!(inside))
                    {
                        inside = true;
                        for (int i = 0; i < clipPolygon.Points.Length; i++)
                        {
                            if (!(subjectPolygon.Contains(clipPolygon[i])))
                            {
                                inside = false;
                                break;
                            }
                        }

                        if (inside)
                        {
                            Point[] points = new Point[clipPolygon.Points.Length];
                            for (int i = 0; i < clipPolygon.Points.Length; i++)
                            {
                                points[i] = clipPolygon.Points[i];
                            }
                            polygons.Add(new Polygon(points));
                        }
                    }
                    else
                    {
                        Point[] points = new Point[subjectPolygon.Points.Length];
                        for (int i = 0; i < subjectPolygon.Points.Length; i++)
                        {
                            points[i] = subjectPolygon.Points[i];
                        }
                        polygons.Add(new Polygon(points));
                    }
                }

                return (polygons);
            }
            #endregion

            #region Private Classes
            private class PolygonPoint
            {
                #region Class Members
                private Point _point;

                private PolygonPoint _next;

                private bool _intersection;

                private bool _visited;

                private int _id;

                private bool _inbound;
                #endregion

                #region Constructors
                internal PolygonPoint(Point point)
                {
                    _point = point;

                    _next = null;
                    _intersection = false;
                    _visited = false;

                    _id = -1;
                    _inbound = false;
                }
                #endregion

                #region Properties
                internal Point Point
                {
                    get { return _point; }
                    set { _point = value; }
                }

                internal PolygonPoint Next
                {
                    get { return _next; }
                    set { _next = value; }
                }

                internal bool IsIntersection
                {
                    get { return _intersection; }
                    set { _intersection = value; }
                }

                internal bool IsVisited
                {
                    get { return _visited; }
                    set { _visited = value; }
                }

                internal int Id
                {
                    get { return _id; }
                    set { _id = value; }
                }

                internal bool IsInbound
                {
                    get { return _inbound; }
                    set { _inbound = value; }
                }
                #endregion
            }
            #endregion
        }
        #endregion

        #region Class Members
        private Point[] _pts;
        #endregion

        #region Constructors
        public Polygon(Point[] points)
        {
            _pts = points;
        }
        #endregion

        #region Properties
        /// <summary>
        /// Number of points the polygon contains.
        /// </summary>
        public int Length
        {
            get { return _pts.Length; }
        }

        /// <summary>
        /// The points of the polygon.
        /// </summary>
        public virtual Point[] Points
        {
            get { return _pts; }
            set { _pts = value; }
        }

        /// <summary>
        /// The area of the polygon.
        /// </summary>
        public double Area
        {
            get
            {
                double area = 0.0;
                for (int i = 0; i < _pts.Length; i++)
                {
                    area += ((_pts[i % _pts.Length].X
                        * _pts[(i + 1) % _pts.Length].Y)
                        - (_pts[(i + 1) % _pts.Length].X
                        * _pts[i % _pts.Length].Y));
                }
                area *= 0.5;

                return (area);
            }
        }

        /// <summary>
        /// An axis-aligned bounding box of the polygon.
        /// </summary>
        public Rect BoundingBox
        {
            get
            {
                double boundsMinX = Points[0].X;
                double boundsMinY = Points[0].Y;
                double boundsMaxX = Points[0].X;
                double boundsMaxY = Points[0].Y;

                for (int i = 0; i < Points.Length; i++)
                {
                    double x = Points[i].X;
                    boundsMinX = Math.Min(boundsMinX, x);
                    boundsMaxX = Math.Max(boundsMaxX, x);

                    double y = Points[i].Y;
                    boundsMinY = Math.Min(boundsMinY, y);
                    boundsMaxY = Math.Max(boundsMaxY, y);
                }
                return (new Rect(boundsMinX, boundsMinY,
                    boundsMaxX - boundsMinX,
                    boundsMaxY - boundsMinY));
            }
        }
        #endregion

        #region Hit Test Methods
        public bool Contains(Point point)
        {
            return (Contains(point.X, point.Y));
        }

        /// <summary>
        /// Check if the polygon contains the given point.
        /// </summary>
        /// <author>Sebastian Boring, Martin Weigel</author>
        /// <param name="x">X coordinate of the point.</param>
        /// <param name="y">Y coordinate of the point.</param>
        /// <returns>True if the polygon contains the given point, otherwise false.</returns>
        public bool Contains(double x, double y)
        {
            if (Points.Length <= 2 || !(BoundingBox.Contains(new Point(x, y))))
            {
                return false;
            }

            // Modification by Martin: Contains corners
            foreach (Point p in Points)
            {
                if (double.Equals(p.X, x) && double.Equals(p.Y, y))
                    return true;
            }

            int hits = 0;

            double lastx = Points[Points.Length - 1].X;
            double lasty = Points[Points.Length - 1].Y;

            double curx, cury;
            for (int i = 0; i < Points.Length; lastx = curx, lasty = cury, i++)
            {
                curx = Points[i].X;
                cury = Points[i].Y;

                if (cury == lasty)
                {
                    continue;
                }

                double leftx;
                if (curx < lastx)
                {
                    if (x >= lastx)
                    {
                        continue;
                    }
                    leftx = curx;
                }
                else
                {
                    if (x >= curx)
                    {
                        continue;
                    }
                    leftx = lastx;
                }

                double test1, test2;
                if (cury < lasty)
                {
                    if (y < cury || y >= lasty)
                    {
                        continue;
                    }
                    if (x < leftx)
                    {
                        hits++;
                        continue;
                    }
                    test1 = x - curx;
                    test2 = y - cury;
                }
                else
                {
                    if (y < lasty || y >= cury)
                    {
                        continue;
                    }
                    if (x < leftx)
                    {
                        hits++;
                        continue;
                    }
                    test1 = x - lastx;
                    test2 = y - lasty;
                }

                if (test1 < (test2 / (lasty - cury) * (lastx - curx)))
                {
                    hits++;
                }
            }

            return ((hits & 1) != 0);
        }

        /// <summary>
        /// Check if polygon contains given polygon.
        /// </summary>
        /// <author>Martin Weigel</author>
        /// <param name="polygon">Given polygon</param>
        /// <returns>True if it contains the polygon, otherwise false.</returns>
        public bool Contains(Polygon polygon)
        {
            foreach (Point p in polygon.Points)
            {
                if (!Contains(p))
                    return false;
            }
            return true;
        }
        #endregion

        #region Intersect Methods
        internal List<LineSegment> GenerateLineSegments()
        {
            List<LineSegment> segments = new List<LineSegment>();
            for (int i = 0; i < Points.Length - 1; i++)
            {
                segments.Add(new LineSegment(this[i], this[i + 1]));
            }
            segments.Add(new LineSegment(this[Points.Length - 1], this[0]));

            return (segments);
        }

        public List<Polygon> IntersectWith(Polygon polygon)
        {
            if (polygon == null
                || !(BoundingBox.IntersectsWith(polygon.BoundingBox)))
            {
                return null;
            }

            List<Polygon> polygons = PolygonClipper.Clip(polygon, this);
            return (polygons);
        }
        #endregion

        #region Other Methods (by Martin)
        private double ZComponent(Point p1, Point p2, Point p3)
        {
            double dx1 = p2.X - p1.X;
            double dy1 = p2.Y - p1.Y;
            double dx2 = p3.X - p2.X;
            double dy2 = p3.Y - p2.Y;
            return dx1 * dy2 - dy1 * dx2;
        }
        private bool SameSign(double d1, double d2)
        {
            return (d1 <= 0 && d2 <= 0) || (d1 >= 0 && d2 >= 0);
        }

        /// <summary>
        /// Check if the polygon is convex. Runtime: O(n)
        /// </summary>
        /// <see cref="http://stackoverflow.com/a/1881201"/>
        /// <author>Martin Weigel</author>
        /// <returns>Return true if the polygon is convex, otherwise false.</returns>
        public bool IsConvex()
        {
            double res1 = ZComponent(this[0], this[1], this[2]);
            for (int i = 1; i <= Length + 2; i++)
            {
                double res2 = ZComponent(this[i], this[i + 1], this[i + 2]);
                if (!SameSign(res1, res2))
                    return false;
            }
            return true;
        }

        /// <summary>
        /// Calculate the centroid of a non-intersecting polygon.
        /// </summary>
        /// <see cref="https://en.wikipedia.org/wiki/Centroid"/>
        /// <author>Martin Weigel</author>
        /// <returns>Centroid of the polygon.</returns>
        public Point Centroid()
        {
            double factor = 1 / (6 * Area);
            double x = 0, y = 0;

            for (int i = 0; i < Length; i++)
            {
                double cross = (this[i].X * this[i + 1].Y - this[i + 1].X * this[i].Y);
                x += (this[i].X + this[i + 1].X) * cross;
                y += (this[i].Y + this[i + 1].Y) * cross;
            }
            return new Point(x * factor, y * factor);
        }

        /// <summary>
        /// Calculates a valid index from the given one. All items are circled by this function to prevent over and underflows.
        /// </summary>
        /// <example>
        ///     Points = [A, B, C]
        ///     IndexAt(1) = 1
        ///     IndexAt(-1) = 2
        ///     IndexAt(4) = 1
        /// </example>
        /// <author>Martin Weigel</author>
        /// <param name="index">Some index</param>
        /// <returns>A valid index for the given index.</returns>
        private int NormalizeIndex(int index)
        {
            // Normalize a positive index
            while (index >= Points.Length)
                index -= Points.Length;
            // Normalize a negative index
            while (-index >= Points.Length)
                index += Points.Length;

            if (index >= 0)
                return index;
            else
                return Points.Length + index;
        }

        /// <summary>
        /// Gets or sets the point at the given index. The index is circulated to avoid over and underflows.
        /// </summary>
        /// <example>
        ///     Points = [A, B, C]
        ///     IndexAt(1) = B
        ///     IndexAt(-1) = C
        ///     IndexAt(4) = B
        /// </example>
        /// <author>Martin Weigel</author>
        /// <param name="index">Index of the point</param>
        /// <returns>The point at the given index.</returns>
        public Point this[int index]
        {
            get { return _pts[NormalizeIndex(index)]; }
            internal set { _pts[NormalizeIndex(index)] = value; }
        }

        /// <summary>
        /// Check if both polygons are the same.
        /// </summary>
        /// <author>Martin Weigel</author>
        /// <param name="polygon">Other polygon for comparison</param>
        /// <returns>If the polygons are the same.</returns>
        public bool IsSame(Polygon polygon)
        {
            if (polygon == null || polygon.Length != Length)
                return false;

            for (int i = 0; i < Length; i++)
            {
                bool match = false;
                for (int j = 0; j < Length; j++)
                {
                    if (this[i] == polygon[j])
                    {
                        match = true;
                        break;
                    }
                }
                if (!match)
                    return false;
            }
            return true;
        }
        #endregion

        #region Overridden Methods (Object)
        public override string ToString()
        {
            string str = "";

            if (_pts != null
                && _pts.Length > 0)
            {
                str += "Polygon [" + _pts.Length + "]: ";
                for (int i = 0; i < _pts.Length; i++)
                {
                    str += "\n\tPoint #" + i + ": {" + _pts[i].X
                        + ", " + _pts[i].Y + "}";
                }
            }
            else
            {
                str += "Polygon [EMPTY]";
            }

            return str;
        }
        #endregion
    }
}
