﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Windows.Media.Media3D;
using ProjectorKit.Elements;
using ProjectorKit.Utilities;

namespace ProjectorKit.World
{
    /// <summary>
    /// The world object is a singleton. It should hold all elements. They are 
    /// automatically displayed and updated in the given WPF Model3DGroup. If 
    /// the given network is a server, all elements are broadcasted to the
    /// clients. Therefore, only one World with a server network should exist.
    /// Make sure to create a world before creating any objects in the 
    /// Relation, Handlers or Tracking namespace.
    /// </summary>
    public abstract class AbstractWorld
    {
        protected ConcurrentDictionary<string, AbstractElement> elements = new ConcurrentDictionary<string, AbstractElement>();
        protected ConcurrentDictionary<AbstractElement, string> nameLookup = new ConcurrentDictionary<AbstractElement, string>();

        #region Properties
        public IDictionary<string, AbstractElement> NameAndElements
        {
            get { return elements; }
        }
        public IEnumerable<AbstractElement> Elements
        {
            get { return elements.Values; }
        }
        public IEnumerable<string> ElementsNames
        {
            get { return elements.Keys; }
        }
        #endregion

        #region Intersection Handling
        public AbstractElement PointingTarget(IPositionable pointer)
        {
            return PointingTarget(pointer.Position, pointer.Direction);
        }
        public AbstractElement PointingTarget(IPositionable pointer, AbstractElement excluding)
        {
            return PointingTarget(pointer.Position, pointer.Direction, excluding);
        }
        public AbstractElement PointingTarget(IPositionable pointer, ICollection<AbstractElement> excluding)
        {
            return PointingTarget(pointer.Position, pointer.Direction, excluding);
        }
        public AbstractElement PointingTarget(Point3D origin, Vector3D direction)
        {
            return PointingTarget(origin, direction, new List<AbstractElement>());
        }
        public AbstractElement PointingTarget(Point3D origin, Vector3D direction, AbstractElement excluding)
        {
            return PointingTarget(origin, direction, new List<AbstractElement>() { excluding });
        }
        public AbstractElement PointingTarget(Point3D origin, Vector3D direction, ICollection<AbstractElement> excluding)
        {
            var targets = PointingTargets(origin, direction, excluding);
            if (targets.Count > 0)
                return targets[0].Value;
            return null;
        }

        public List<KeyValuePair<double, AbstractElement>> PointingTargets(IPositionable pointer)
        {
            return PointingTargets(pointer.Position, pointer.Direction, new List<AbstractElement>());
        }
        public List<KeyValuePair<double, AbstractElement>> PointingTargets(IPositionable pointer, ICollection<AbstractElement> excluding)
        {
            return PointingTargets(pointer.Position, pointer.Direction, excluding);
        }
        public List<KeyValuePair<double, AbstractElement>> PointingTargets(IPositionable pointer, AbstractElement excluding)
        {
            return PointingTargets(pointer.Position, pointer.Direction, new List<AbstractElement>() { excluding });
        }
        public List<KeyValuePair<double, AbstractElement>> PointingTargets(Point3D origin, Vector3D direction)
        {
            return PointingTargets(origin, direction, new List<AbstractElement>());
        }
        public List<KeyValuePair<double, AbstractElement>> PointingTargets(Point3D origin, Vector3D direction, AbstractElement excluding)
        {
            return PointingTargets(origin, direction, new List<AbstractElement>() { excluding });
        }
        public List<KeyValuePair<double, AbstractElement>> PointingTargets(Point3D origin, Vector3D direction, ICollection<AbstractElement> excluding)
        {
            var targets = new List<KeyValuePair<double, AbstractElement>>();
            foreach (AbstractElement e in elements.Values)
            {
                if (!excluding.Contains(e))
                {
                    Point3D? hitpoint = e.Intersection(origin, direction);
                    if (hitpoint != null)
                    {
                        double newDistance = origin.DistanceTo((Point3D)hitpoint);
                        targets.Add(new KeyValuePair<double, AbstractElement>(newDistance, e));
                    }
                }
            }

            targets.Sort(Compare);
            return targets;
        }
        private int Compare(KeyValuePair<double, AbstractElement> a, KeyValuePair<double, AbstractElement> b)
        {
            return a.Key.CompareTo(b.Key);
        }
        #endregion

        #region InternalElementHandling
        protected abstract void ElementMeshRefresh(AbstractElement element);

        protected void RegisterElement(string name, AbstractElement element)
        {
            if (elements.Values.Contains(element))
                throw new ArgumentException("The element is already registered!");
            if (elements.ContainsKey(name))
                throw new ArgumentException("An element with the same name is already registered!");
            
            element.Updated += ElementMeshRefresh;
            elements.TryAdd(name, element);
            nameLookup.TryAdd(element, name);
        }

        protected void UnregisterElement(string name)
        {
            AbstractElement element;
            if (elements.TryGetValue(name, out element))
            {
                elements[name].Updated -= ElementMeshRefresh;
                nameLookup.TryRemove(elements[name], out name);
                elements.TryRemove(name, out element);
            }
            else
                throw new ArgumentException("The element does not exist!");
        }
        #endregion

        #region PublicElementHandling
        public bool TryGetElement(string name, out AbstractElement element)
        {
            return elements.TryGetValue(name, out element);
        }
        public bool TryGetName(AbstractElement element, out string name)
        {
            return nameLookup.TryGetValue(element, out name);
        }

        public bool Contains(string name)
        {
            return elements.ContainsKey(name);
        }

        public bool Contains(AbstractElement element)
        {
            return nameLookup.ContainsKey(element);
        }

        public AbstractElement Get(string name)
        {
            AbstractElement element;
            return TryGetElement(name, out element) ? element : null;
        }

        public string GetNameOf(AbstractElement element)
        {
            string name;

            if (nameLookup.TryGetValue(element, out name))
                return name;
            else
                return null;
        }
        #endregion
    }
}
