﻿using System;
using System.Collections.Generic;
using GroupLab.iNetwork;
using ProjectorKit.Elements;
using ProjectorKit.Filter;
using ProjectorKit.Network;

namespace ProjectorKit.World
{
    public class ServerWorld : AbstractWorld
    {
        #region Singleton
        protected static ServerWorld instance;
        public static ServerWorld Instance
        {
            get
            {
                if (instance == null)
                    throw new Exception("No world exist yet!");
                return instance;
            }
        }
        #endregion

        private BufferServer server;
        private IFilterFactory defaultElementFilterFactory = new NoFilterFactory();
        #region Properties
        public IFilterFactory DefaultElementFilter
        {
            get { return defaultElementFilterFactory; }
            set
            {
                defaultElementFilterFactory = value;
                foreach (AbstractElement e in Elements)
                    e.Filter = defaultElementFilterFactory.NewFilter();
            }
        }
        #endregion

        public ServerWorld(BufferServer server)
        {
            if (instance != null)
                throw new Exception("Only one world can exist in an application.");

            instance = this;
            this.server = server;
        }

        public void ProjectElement(AbstractElement2D element, Projector projector, int x, int y, int width, int height)
        {
            element.ProjectWith(this, projector, GetNameOf(projector), x, y, width, height);
            Add(element);
        }

        public void Focus(AbstractElement element)
        {
            Message msg = new Message(WorldMessageNames.Focus);
            msg.AddField("id", GetNameOf(element));
            server.Broadcast(msg);
        }

        private void SendNetworkUpdate(AbstractElement element)
        {
            if (server != null)
                server.Broadcast(UpdateMessageForElement(element));
        }

        #region PublicElementHandling
        public void Add(AbstractElement element)
        {
            if(!Contains(element))
                Add("UNNAMED_" + element.GetHashCode().ToString(), element);
        }
        public void Add(IEnumerable<AbstractElement> elements)
        {
            foreach (AbstractElement element in elements)
                Add(element);
        }

        public void Add(string name, AbstractElement element)
        {
            element.Updated += SendNetworkUpdate;
            RegisterElement(name, element);

            // Set default filter to element if no filter was explicitly set
            if (element.Filter is NoFilter)
                element.Filter = defaultElementFilterFactory.NewFilter();

            SendNetworkUpdate(element);
        }

        public void Remove(AbstractElement element)
        {
            string name;
            if (TryGetName(element, out name))
            {
                elements[name].Updated -= SendNetworkUpdate;
                UnregisterElement(name);

                // Remove element from network
                if (server != null)
                {
                    Message msg = new Message(WorldMessageNames.Remove);
                    msg.AddField("id", name);
                    server.Broadcast(msg);
                }
            }
        }
        #endregion

        public Message UpdateMessageForElement(AbstractElement element)
        {
            Message msg = element.ToMessage(WorldMessageNames.Update);
            msg.AddField("id", GetNameOf(element));
            return msg;
        }

        protected override void ElementMeshRefresh(AbstractElement element) {}
    }
}
